import os
import numpy as np
import xml.etree.ElementTree as et
import msm.utilities as utl
# import msmsg as msg


# ====================================================================
# VABS

def readVABSOutLocal(fn_vabs_in):

    return


def writeVABSInLocal(fn_vabs_in, gebtin_name):
    gebtin = readGEBTIn(gebtin_name)
    gebtout_ptr, gebtout_mbr = readGEBTOut(gebtin_name, gebtin)

    # if gebtin['analysis'] == 0:
    step = -1
    point = 1
    member = 1
    element = 1
    if step > 0:
        step -= 1
    if point > 0:
        point -= 1
    if member > 0:
        member -= 1
    if element > 0:
        element -= 1
    point_result = gebtout_ptr[step][point]
    element_result = gebtout_mbr[step][member][element]

    c = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]

    vabsin_name_homo = fn_vabs_in + '.homo'
    try:
        os.rename(fn_vabs_in, vabsin_name_homo)
    except WindowsError:
        pass
    with open(fn_vabs_in, 'w') as fout:
        # Change the recover flag
        with open(vabsin_name_homo, 'r') as fin:
            changed = False
            for line in fin:
                if (not changed) and (len(line.split()) == 3):
                    line = map(int, line.split())
                    beam_model = line[0]
                    line[1] = 2
                    fout.write('{0[0]:4d}{0[1]:4d}{0[2]:4d}\n'.format(line))
                    changed = True
                else:
                    fout.write(line)
        # Append global results
        fout.write('\n')
        # u1 u2 u3
        fout.write('{0[0]:16.7E}{0[1]:16.7E}{0[2]:16.7E}\n'.format(
            element_result[:3]))
        fout.write('\n')
        # C11 C12 C13
        # C21 C22 C23
        # C31 C32 C33
        fout.write('{0[0]:16.7E}{0[1]:16.7E}{0[2]:16.7E}\n'.format(c[0]))
        fout.write('{0[0]:16.7E}{0[1]:16.7E}{0[2]:16.7E}\n'.format(c[1]))
        fout.write('{0[0]:16.7E}{0[1]:16.7E}{0[2]:16.7E}\n'.format(c[2]))
        fout.write('\n')
        # F1 M1 M2 M3
        fout.write('{0:16.7E}{1[0]:16.7E}{1[1]:16.7E}{1[2]:16.7E}\n'.format(
            element_result[9], element_result[12:]))
        if beam_model == 1:  # Timoshenko model
            # F2 F3
            fout.write('{0[0]:16.7E}{0[1]:16.7E}\n'.format(
                element_result[10:12]))
            fout.write('\n')
            # f1    f2    f3    m1    m2    m3
            # f1'   f2'   f3'   m1'   m2'   m3'
            # f1''  f2''  f3''  m1''  m2''  m3''
            # f1""" f2""" f3""" m1""" m2""" m3"""
            fout.write('{0:16.7E}{1:16.7E}{2:16.7E}{3:16.7E}{4:16.7E}{5:16.7E}\n'.format(
                0, 0, 0, 0, 0, 0))
            fout.write('{0:16.7E}{1:16.7E}{2:16.7E}{3:16.7E}{4:16.7E}{5:16.7E}\n'.format(
                0, 0, 0, 0, 0, 0))
            fout.write('{0:16.7E}{1:16.7E}{2:16.7E}{3:16.7E}{4:16.7E}{5:16.7E}\n'.format(
                0, 0, 0, 0, 0, 0))
            fout.write('{0:16.7E}{1:16.7E}{2:16.7E}{3:16.7E}{4:16.7E}{5:16.7E}\n'.format(
                0, 0, 0, 0, 0, 0))
        fout.write('\n')

    return


# ====================================================================
# SC

def preSC(fn_xml):
    # debug = True
    print('running presc...')

    # ====================================================================
    # Parse XML parametric inputs
    tree = et.parse(fn_xml)
    sg = tree.getroot()
    name = sg.get('name')
    nsg = int(sg.find('dimension').text)
    macro_dim = int(sg.find('macro').text)

    fn_sc = name + '.sc'

    # Materials and Laminae
    dict_type = {'isotropic': 0, 'orthotropic': 1, 'anisotropic': 2}
    elastic_label = {
        'isotropic': (('e', 'nu'),),
        'orthotropic': (
            ('e1', 'e2', 'e3'),
            ('g12', 'g13', 'g23'),
            ('nu12', 'nu13', 'nu23')
        ),
        'anisotropic': (
            ('c11', 'c12', 'c13', 'c14', 'c15', 'c16'),
            ('c22', 'c23', 'c24', 'c25', 'c26'),
            ('c33', 'c34', 'c35', 'c36'),
            ('c44', 'c45', 'c46'),
            ('c55', 'c56'),
            ('c66',)
        )
    }
    d_material = {}
    d_lamina = {}
    se_materials = sg.find('materials')
    for mtr in se_materials.findall('material'):
        mn = mtr.get('name').strip()
        mt = mtr.get('type').strip()
        md = float(mtr.find('density').text.strip())
        label = elastic_label[mt]
        se_elastic = mtr.find('elastic')
        me = {}
        for i in range(len(label)):
            for j in range(len(label[i])):
                me[label[i][j]] = float(se_elastic.find(label[i][j]).text.strip())
        se_strength = mtr.find('strength')
        d_material[mn] = {'type': dict_type[mt], 'density': md, 'elastic': me}
    for lmn in se_materials.findall('lamina'):
        ln = lmn.get('name').strip()
        lm = lmn.find('material').text.strip()
        lt = float(lmn.find('thickness').text.strip())
        d_lamina[ln] = {'material': lm, 'thickness': lt}

    # Head
    analysis = int(sg.find('analysis').text)
    macro_model = int(sg.find('model').text)  # model (0: classical, 1: shear refined)
    elem_flag = int(sg.find('elem_flag').text)
    trans_flag = int(sg.find('trans_flag').text)
    temp_flag = int(sg.find('temp_flag').text)

    # Layup
    ld_layer = []  # List of material, thickness, angle
    l_material_used = []
    layup = sg.find('layup')
    method = 'll'  # ll: layer list, ss: stacking sequence, pp: ply percentage
    if layup.get('method') is not None:
        method = layup.get('method')
    if method == 'll':
        # layer list
        for lyr in layup.findall('layer'):
            ln = lyr.get('lamina').strip()
            lm = d_lamina[ln]['material']
            if lm not in l_material_used:
                l_material_used.append(lm)
                d_material[lm]['id'] = len(l_material_used)
            lt = d_lamina[ln]['thickness']
            lo = float(lyr.text.strip())
            ld_layer.append({'material': lm, 'thickness': lt, 'angle': lo})
    elif method == 'ss':
        # stacking sequence
        ln = layup.find('lamina').text.strip()
        lm = d_lamina[ln]['material']
        l_material_used.append(lm)
        d_material[lm]['id'] = len(l_material_used)
        lt = d_lamina[ln]['thickness']
        layup_code = layup.find('code').text.strip()
        code = utl.parseLayupCode(layup_code)
        for lo in code:
            ld_layer.append({'material': lm, 'thickness': lt, 'angle': lo})
    elif method == 'pp':
        # ply percentage
        pass
    # print code
    nmate = len(l_material_used)

    # Rearrange
    tt = 0.0  # total thickness
    d_layertype_used = {}
    for lyr in ld_layer:
        tt = tt + lyr['thickness']
        layertype_id = 0
        for k, v in d_layertype_used.items():
            if (v['material'] == lyr['material']) and (v['angle'] == lyr['angle']):
                layertype_id = k
        if layertype_id == 0:
            # new layer type
            layertype_id = len(d_layertype_used) + 1
            d_layertype_used[layertype_id] = {
                'material': lyr['material'], 'angle': lyr['angle']
            }
        lyr['layertype'] = layertype_id

    nlayer = len(d_layertype_used)

    # Create elements
    pan = 0.0
    if sg.find('translate') is not None:
        pan = float(sg.find('translate').text.strip())
    nep = 1  # By default, one element per ply
    gms = float(sg.find('global_mesh_size').text.strip())
    ens = int(sg.find('element_nodes').text.strip())
    # nep = int(round(thickness / gms))  # number of elements per ply (rounded)
    # print nep

    nply = len(ld_layer)
    # tthk = thickness * nply
    ht = tt / 2.0
    # print tthk

    nodes_major = np.array([-ht, ])
    for lyr in ld_layer:
        t = lyr['thickness']
        ne = int(round(t / gms))
        lyr['nelem'] = ne
        ns = np.linspace(nodes_major[-1], nodes_major[-1] + t, ne+1)
        nodes_major = np.hstack((nodes_major, ns[1:]))

    # nodes_major = [0.0, ]
    # nodes_major = np.linspace(-y3mid, y3mid, nep * nply + 1)
    # print nodes_major
    nelem = len(nodes_major) - 1

    nodes_q1 = []
    nodes_q2 = []
    nodes_q3 = []

    if macro_dim == 2:
        # As the SC manual suggested,
        # Use 5-node elements for 2D plate/shell models
        for i in range(nelem):
            nodes_q2.append((nodes_major[i] + nodes_major[i + 1]) / 2.)
            nodes_q1.append((nodes_major[i] + nodes_q2[i]) / 2.)
            nodes_q3.append((nodes_q2[i] + nodes_major[i + 1]) / 2.)
    # print nodes_major
    # print nodes_q1
    # print nodes_q2
    # print nodes_q3
    nnode = len(nodes_major) + len(nodes_q1 + nodes_q2 + nodes_q3)
    # nelem = len(nodes_q2)

    # ====================================================================
    # Write SwiftComp input file

    # string format
    sfi = '{:8d}'
    sff = '{:16.6E}'

    with open(fn_sc, 'w') as fout:
        # Extra inputs for dimensionally reducible structures
        if (macro_dim == 1) or (macro_dim == 2):
            # model (0: classical, 1: shear refined)
            fout.write((sfi + '\n').format(macro_model))
            if macro_dim == 1:  # beam
                # initial twist/curvatures
                fout.write((sff * 3 + '\n').format(0., 0., 0.))
                # oblique cross section
                fout.write((sff * 2 + '\n').format(1., 0.))
            elif macro_dim == 2:  # shell
                # initial twist/curvatures
                fout.write((sff * 2 + '\n').format(0., 0.))
            fout.write('\n')

        # Head
        fout.write((sfi * 4 + '\n').format(analysis,
                                           elem_flag, trans_flag, temp_flag))
        fout.write((sfi * 6 + '\n').format(nsg,
                                           nnode, nelem, nmate, 0, nlayer))

        # Nodal coordinates
        fout.write('\n')
        nid = 1
        nids_major = []
        nids_q1 = []
        nids_q2 = []
        nids_q3 = []
        for y3 in nodes_major:
            fout.write((sfi + sff + '\n').format(nid, y3 + pan))
            nids_major.append(nid)
            nid += 1
        if macro_dim == 2:
            # As the SC manual suggested,
            # Use 5-node elements for 2D plate/shell models
            for y3 in nodes_q2:
                fout.write((sfi + sff + '\n').format(nid, y3 + pan))
                nids_q2.append(nid)
                nid += 1
            for y3 in nodes_q1:
                fout.write((sfi + sff + '\n').format(nid, y3 + pan))
                nids_q1.append(nid)
                nid += 1
            for y3 in nodes_q3:
                fout.write((sfi + sff + '\n').format(nid, y3 + pan))
                nids_q3.append(nid)
                nid += 1

        # Element connectivities
        fout.write('\n')
        eid = 0
        for i, lyr in enumerate(ld_layer):
            for j in range(lyr['nelem']):
                eid += 1
                line = [
                    eid, lyr['layertype'], 
                    nids_major[eid-1], nids_major[eid], 0, 0, 0
                ]
                if macro_dim == 2:
                    # As the SC manual suggested,
                    # Use 5-node elements for 2D plate/shell models
                    line[-3:] = [nids_q1[eid-1], nids_q3[eid-1], nids_q2[eid-1]]
                utl.writeFormatIntegers(fout, line)

        # Layer types
        fout.write('\n')
        # for lid in range(nlayer):
        #     fout.write((sfi * 2 + sff + '\n').format(lid +
        #                                              1, 1, angle_used[lid]))
        for k, v in d_layertype_used.items():
            mid = d_material[v['material']]['id']
            angle = v['angle']
            fout.write((sfi * 2 + sff + '\n').format(k, mid, angle))

        # Material
        fout.write('\n')
        for mn in l_material_used:
            mtr = d_material[mn]
            fout.write((sfi * 3 + '\n').format(mtr['id'], mtr['type'], 1))
            fout.write((sff * 2 + '\n').format(0., mtr['density']))
            for i in range(len(label)):
                for j in range(len(label[i])):
                    fout.write(sff.format(mtr['elastic'][label[i][j]]))
                fout.write('\n')

        # Omega
        fout.write('\n')
        if macro_dim == 3:
            fout.write((sff + '\n').format(tt))
        elif macro_dim == 2:
            fout.write((sff + '\n').format(1.0))

        fout.write('\n')

    return


def readSCOutFailure(fn_swiftcomp_in, failure_analysis):
    fn_swiftcomp_in += '.fi'
    lines = []
    kw_index = {}
    results = []
    with open(fn_swiftcomp_in, 'r') as fin:
        for i, line in enumerate(fin):
            lines.append(line)
            if failure_analysis == 'f':
                # initial failue strength
                if 'Initial Failure Strengths' in line:
                    kw_index['Initial Failure Strengths'] = i
            elif failure_analysis == 'fe':
                # initial failure envelope
                pass
            elif failure_analysis == 'fi':
                # initial failure indices and strength ratios
                pass

    if failure_analysis == 'f':
        # initial failure strength
        try:
            index = kw_index['Initial Failure Strengths']
            for line in lines[index + 2:index + 8]:
                line = line.strip().split()
                results.append(map(float, line[:2]))
        except KeyError:
            print('No initial failure strength found.')
    elif failure_analysis == 'fe':
        # initial failure envelope
        for line in lines:
            line = line.strip().split()
            results.append([int(line[0]), float(line[1]), float(line[2])])
    elif failure_analysis == 'fi':
        # initial failure indices and strength ratios
        for line in lines:
            line = line.strip().split()
            results.append([int(line[0]), float(line[1]), float(line[2])])

    return results


# ====================================================================
# GEBT

def readGEBTIn(gebtInputName):
    results = {}
    lines = []

    with open(gebtInputName, 'r') as fin:
        for ln, line in enumerate(fin):
            line = line.strip()
            if line == '':
                continue
            else:
                lines.append(line.split('#')[0].strip())

    li = 0
    line1 = map(int, lines[li].split())
    results['analysis'] = line1[0]
    results['niter'] = line1[1]
    results['nstep'] = line1[2]
    if results['analysis'] == 0:
        li += 1
    else:
        li += 5
        if results['analysis'] == 3:
            results['nev'] = int(lines[li])
            li += 1
    line2 = map(int, lines[li].split())
    results['nkp'] = line2[0]
    results['nmemb'] = line2[1]
    results['ncondpt'] = line2[2]
    results['nmate'] = line2[3]
    results['nframe'] = line2[4]
    results['ncondmb'] = line2[5]
    results['ndistr'] = line2[6]
    results['ntimefun'] = line2[7]
    results['ncurv'] = line2[8]
    li += 1
    # Read keypoints
    kps = []
    for i in range(li, li + results['nkp']):
        kps.append(map(float, lines[i].split()[1:]))
    results['points'] = kps
    li += results['nkp']
    # Read members
    membs = []
    for i in range(li, li + results['nmemb']):
        membs.append(map(int, lines[i].split()[1:]))
    results['members'] = membs
    li += results['nmemb']

    return results


# --------------------------------------------------------------------

def readGEBTOutNode(lines):
    out = []
    for l in lines:
        out += map(float, l.strip().split())
    return out


def readGEBTOut(gebtInputName, inputs={}):

    if len(inputs) == 0:
        inputs = readGEBTIn(gebtInputName)

    flag_analysis = inputs['analysis']
    nstep = inputs['nstep']
    nkp = inputs['nkp']
    nmemb = inputs['nmemb']

    members_in = inputs['members']

    gebtInputName += '.out'

    results = []  # for all steps
    points_results = []
    members_results = []

    # lines = []

    with open(gebtInputName, 'r') as fin:
        all_lines = fin.readlines()

    i = 0
    sid = 0  # at least one step
    mid = 0
    while i < len(all_lines):
        line = all_lines[i].strip()
        if (line == '') or ('===' in line) or ('---' in line):
            i += 1
            continue
        elif ('Step' in line) or ((nstep == 1) and (sid == 0)):
            sid += 1  # step number
            results_step = [[], []]  # point results, member results
            mid = 0
            i += 1
            continue
        elif 'Point' in line:
            # Read point result
            out = readGEBTOutNode(all_lines[i + 2:i + 5])
            i += 5
            results_step[0].append(out)
            # points_results.append(out)
            continue
        elif 'Member' in line:
            # Read member result
            mid += 1
            nelem = members_in[mid - 1][-2]  # number of divisions
            i += 2
            member_out = []
            for j in range(nelem):
                if flag_analysis == 0:
                    # Static
                    out = readGEBTOutNode(all_lines[i:i + 3])
                    i += 4
                else:
                    # Dynamic
                    out = readGEBTOutNode(all_lines[i:i + 4])
                    i += 5
                member_out.append(out)
            results_step[1].append(member_out)
            # members_results.append(member_out)
            if mid == nmemb:
                results.append(results_step)
            continue
        i += 1

    return results


# --------------------------------------------------------------------

def readGEBTOutEigen(gebtin_name, gebtin={}):

    if len(gebtin) == 0:
        gebtin = readGEBTIn(gebtin_name)
    gebtout_name = gebtin_name + '.out'

    nkp = gebtin['nkp']
    nmemb = gebtin['nmemb']
    members_in = gebtin['members']
    # print 'nmemb =', nmemb

    eva = []  # Eigenvalues: [[eva11, eva12], [eva21, eva22], ...]
    eve = []  # Eigenvectors: [eve1, eve2, ...]
    # evei = [p1, m11, m12, ..., m1n, p2]
    # p1/m1i = [x1, x2, x3, t1, t2, t3]

    with open(gebtout_name, 'r') as fin:
        all_lines = fin.readlines()

    eig = 0
    i = 0
    block = ''
    while i < len(all_lines):
        # print i
        if all_lines[i].strip() == '':
            i += 1
            continue
        if 'Eigenvalue' in all_lines[i]:
            eig += 1
            # print 'eigenvalue:', eig
            eva.append(map(float, all_lines[i + 1].split()))
            evei = [[], []]
            # store all Point results for an Eigenvalue temporarily
            # evei_p = []  # = [[x11, x12, x13, u11, u12, u13, t11, t12, t13], ...]
            # store all Member results for an Eigenvalue temporarily
            # evei_m = []  # = [[[x11, x12, x13, u11, u12, u13, t11, t12, t13], [], ...], ...]
            mid = 0  # Member id
            pid = 0
            i += 2
            continue
        if eig > 0:
            if 'Point' in all_lines[i]:
                pid += 1
                out = readGEBTOutNode(all_lines[i + 2:i + 5])
                # pr = utl.textToMatrix(all_lines[i + 2:i + 5])  # point results
                # x, u, t = pr[0], pr[1][:3], pr[1][3:]
                # [x11, x12, x13, u11, u12, u13, t11, t12, t13]
                # eve_p.append(pr[0] + pr[1])
                i += 5
                evei[0].append(out)
                continue
            elif 'Member' in all_lines[i]:
                # Read member result
                mid += 1
                nelem = members_in[mid - 1][-2]  # number of divisions
                i += 2
                evei_m = []
                for j in range(nelem):
                    out = readGEBTOutNode(all_lines[i:i + 4])
                    i += 5
                    evei_m.append(out)
                evei[1].append(evei_m)
                # members_results.append(member_out)
                if mid == nmemb:
                    eve.append(evei)
                continue
        i += 1

    results = [eva, eve]
    return results
